@@ -1,15 +1,17 @@
 #!/usr/bin/env python3
 """An open library for simplified-traditional Chinese text conversion.
 """
-import sys, os
+import sys
+import os
 import re
+import glob
 import argparse
 import json
-from glob import glob
 import html
 from collections import namedtuple, OrderedDict
+import math
 
-__version__ = '0.15.0'
+__version__ = '0.16.0'
 __author__ = 'Danny Lin'
 __author_email__ = 'danny0838@gmail.com'
 __homepage__ = 'https://github.com/danny0838/sts-lib'
@@ -151,21 +153,39 @@ class Unicode():
 class StsDict(OrderedDict):
     """Base class of an STS dictionary.
 
-    This class is for child classes to implement on and not intended to be used
-    directly.
+    This class is for child classes to implement on and not intended to be
+    used directly.
+
+    Supported serializations are:
+    - plain-dict: which is a text file that looks like:
+
+          key1 <tab> value1-1 [<space> value1-2 <space> ...]
+          key2 <tab> value2-1 [<space> value2-2 <space> ...]
+          ...
+
+    - JSON: which is dumped from the internal data structure.
+
+    NOTE: The internal data structure may vary across subclasses. As a result,
+    loadjson(), dumpjson(), and most native dict methods like __init__(),
+    __iter__(), items(), keys(), and values() work on their own format.
+
+    To safely create a subclass from a dict-compatible data format, do
+    something like:
+
+        SubClass().add_dict(StsDict(...))
+
+    Or use load() or load_filegroups() to load from a plain-dict file.
     """
     def iter(self):
         """Get a generator of key-values pairs.
         """
-        for key, values in self.items():
-            yield key, values
+        yield from self.items()
 
     def add(self, key, values, skip_check=False):
         """Add a key-values pair to this dictionary.
 
         Args:
-            values: a string or a list of values.
-                ([value1, value2, value3, ...])
+            values: a string or a list of strings.
             skip_check: True to skip checking duplicated values.
         """
         values = [values] if isinstance(values, str) else values
@@ -177,7 +197,7 @@ class StsDict(OrderedDict):
         return self
 
     def add_dict(self, stsdict, skip_check=False):
-        """Add all key-values pairs from another dictionary.
+        """Add all key-values pairs from another StsDict.
 
         Args:
             skip_check: True to skip checking duplicated values.
@@ -187,7 +207,7 @@ class StsDict(OrderedDict):
         return self
 
     def load(self, *files):
-        """Add key-values pairs from file(s) to the dictionary.
+        """Add all key-values pairs from plain-dict file(s).
         """
         for file in files:
             with open(file, "r", encoding="UTF-8") as f:
@@ -200,7 +220,7 @@ class StsDict(OrderedDict):
         return self
 
     def dump(self, file=None, sort=False):
-        """Dump key-values pairs to a text file.
+        """Dump key-values pairs to a plain-dict file.
 
         Args:
             file: path of file to save. Use stdout if None.
@@ -216,6 +236,8 @@ class StsDict(OrderedDict):
     def loadjson(self, file):
         """Load from a JSON file.
 
+        NOTE: The input data format may vary across subclasses.
+
         Returns:
             a new object with the same class.
         """
@@ -227,6 +249,8 @@ class StsDict(OrderedDict):
     def dumpjson(self, file=None, indent=None, sort=False):
         """Dump key-values pairs to a JSON file.
 
+        NOTE: The output data format may vary across subclasses.
+
         Args:
             file: path of file to save. Use stdout if None.
             indent: indent the output with a specified integer.
@@ -248,7 +272,7 @@ class StsDict(OrderedDict):
             print(f'{key} => {" ".join(values)}')
 
     def find(self, keyword, from_keys=True, from_values=True, exact=False):
-        """Search for keyword from all values.
+        """Search for keyword from the dictionary.
 
         Args:
             from_keys: True to search from keys.
@@ -284,21 +308,21 @@ class StsDict(OrderedDict):
                 stsdict.add(value, key)
         return stsdict
 
-    def load_fileslist(self, fileslist):
-        """Load key-values pairs from a list of filelists.
-
-        Returns:
-            a new object with the same class.
+    def load_filegroups(self, filegroups):
+        """Load key-values pairs from a list of plain-dict files.
 
         Args:
-            fileslist: a list of dictionary lists.
+            filegroups: a list of plain-dict files.
             [
                 [file1-1, file1-2, ...]
                 [file2-1, file2-2, file2-3, ...]
                 [file3-1, file3-2, file3-3, ...]
             ]
+
+        Returns:
+            a new object with the same class.
         """
-        stsdicts = [self.__class__().load(*files) for files in fileslist]
+        stsdicts = [self.__class__().load(*files) for files in filegroups]
         newstsdict = self
         while len(stsdicts):
             newstsdict = newstsdict.join(stsdicts.pop(0))
@@ -370,7 +394,7 @@ class StsDict(OrderedDict):
         dict_.add_dict(stsdict)
         return dict_
 
-    def match(self, parts, pos):
+    def match(self, parts, pos, maxlen=math.inf):
         """Match a unicode composite at pos.
 
         Args:
@@ -383,7 +407,7 @@ class StsDict(OrderedDict):
                 parts if isinstance(parts, list) else
                 list(parts))
         i = max(len(Unicode.split(key)) for key in self)
-        i = min(i, len(parts) - pos)
+        i = min(i, min(len(parts), maxlen) - pos)
         while i >= 1:
             end = pos + i
             current = "".join(parts[pos:end])
@@ -457,7 +481,7 @@ class StsDict(OrderedDict):
         while len(queue):
             (parts, matched, nextindex) = data = queue.pop(0)
             if nextindex < len(parts):
-                queue += self._apply_enum_sub(data, include_short=include_short, include_self=include_self)
+                self._apply_enum_sub(queue, data, include_short=include_short, include_self=include_self)
             elif matched > 0:
                 results["".join(parts)] = True
 
@@ -468,73 +492,70 @@ class StsDict(OrderedDict):
 
         return results
 
-    def _apply_enum_sub(self, data, include_short=False, include_self=False):
+    def _apply_enum_sub(self, queue, data, include_short=False, include_self=False):
         """Helper function of apply_enum
 
         Args:
             data: (parts, matched, index)
-
-        Returns:
-            [(value, matched, nextindex), ...]
         """
         (parts, matched, index) = data
-        results = []
+        match = self.match(parts, index)
+
+        # add atomic stepping if no match
+        if match is None:
+            queue.append((parts, matched, index + 1))
+            return
+
         has_atomic_match = False
-        if index < len(parts):
-            match = self.match(parts, index)
+        if match.end - index == 1:
+            has_atomic_match = True
+        if include_self and match.conv.key not in match.conv.values:
+            match.conv.values.append(match.conv.key)
+        for value in match.conv.values:
+            result = parts[:index] + [value] + parts[match.end:]
+            queue.append((result, matched + 1, index + 1))
+
+        if not include_short:
+            return
+
+        # add shorter matches
+        for i in range(match.end - match.start - 1, 0, -1):
+            match = self.match(parts, index, maxlen=index + i)
             if match is not None:
                 if match.end - index == 1:
                     has_atomic_match = True
+                if include_self and match.conv.key not in match.conv.values:
+                    match.conv.values.append(match.conv.key)
                 for value in match.conv.values:
                     result = parts[:index] + [value] + parts[match.end:]
-                    results.append((result, matched + 1, index + 1))
-                if include_self and match.conv.key not in match.conv.values:
-                    result = parts[:index] + [match.conv.key] + parts[match.end:]
-                    results.append((result, matched + 1, index + 1))
+                    queue.append((result, matched + 1, index + 1))
 
-                if not include_short:
-                    return results
-
-                # add shorter matches
-                for i in range(match.end - match.start - 1, 0, -1):
-                    match = self.match(parts[:index + i], index)
-                    if match is not None:
-                        if match.end - index == 1:
-                            has_atomic_match = True
-                        for value in match.conv.values:
-                            result = parts[:index] + [value] + parts[match.end:]
-                            results.append((result, matched + 1, index + 1))
-                        if include_self and match.conv.key not in match.conv.values:
-                            result = parts[:index] + [match.conv.key] + parts[match.end:]
-                            results.append((result, matched + 1, index + 1))
-
-                # add atomic match (length = 1) case
-                #
-                # e.g.
-                # table: 采信 => 採信, 信息 => 訊息
-                # text: ["采", "信", "息"]
-                #
-                # We get a match ["采", "信", "息"] but no atomic match available.
-                #                 ^^^^^^^^^^
-                #
-                # Add ["采", "信", "息"] so that "采訊息" is not missed.
-                #             ^
-                if not has_atomic_match:
-                    results.append((parts, matched, index + 1))
-
-        if len(results) == 0:
-            results.append((parts, matched, index + 1))
-
-        return results
+        # add atomic stepping (length = 1) case
+        #
+        # e.g.
+        # table: 采信 => 採信, 信息 => 訊息
+        # text: ["采", "信", "息"]
+        #
+        # We get a match ["采", "信", "息"] but no atomic match available.
+        #                 ^^^^^^^^^^
+        #
+        # Add ["采", "信", "息"] so that "采訊息" is not missed.
+        #             ^
+        if not has_atomic_match:
+            queue.append((parts, matched, index + 1))
 
 class Table(StsDict):
-    """An STS dictionary with hash format.
+    """A cache-boosted STS dictionary.
 
-    This class implements a cache and a head-char checking mechanism to improve
-    performance on text conversion than base StsDict. The cache will not update
-    after generated, and therefore the dictionary should not be modified after
-    performing a text matching or conversion (match, apply, apply_enum, join,
-    etc.) to avoid an unexpected behavior.
+    This class implements a cache and a head-char checking mechanism to
+    improve performance on text conversion than base StsDict.
+
+    The internal data format is same as base StsDict.
+
+    NOTE: The cache will not update after generated, and therefore the
+    dictionary should not be modified after performing a text matching or
+    conversion (match, apply, apply_enum, etc.) to avoid an unexpected
+    behavior.
     """
     @lazyprop
     def key_maxlen(self):
@@ -548,7 +569,7 @@ class Table(StsDict):
         """
         return set(key[0] for key in self)
 
-    def match(self, parts, pos):
+    def match(self, parts, pos, maxlen=math.inf):
         """Match a unicode composite at pos.
 
         Args:
@@ -562,7 +583,7 @@ class Table(StsDict):
                 list(parts))
         if parts[pos][0] in self.key_headchars:
             i = self.key_maxlen
-            i = min(i, len(parts) - pos)
+            i = min(i, min(len(parts), maxlen) - pos)
             while i >= 1:
                 end = pos + i
                 current = "".join(parts[pos:end])
@@ -577,12 +598,14 @@ class Trie(StsDict):
 
     Compared with hash, trie is faster for applying conversion,
     but takes more space and is slower to construct.
+
+    NOTE: The internal data format is different from base StsDict.
     """
     def iter(self):
         """Get a generator of key-values pairs.
         """
         def recurse(trie):
-            for key in trie.keys():
+            for key in trie:
                 if key == "":
                     yield ''.join(keystack), trie[key]
                 else:
@@ -600,8 +623,7 @@ class Trie(StsDict):
         """Add a key-values pair to this dictionary.
 
         Args:
-            values: a string or a list of values.
-                ([value1, value2, value3, ...])
+            values: a string or a list of strings.
             skip_check: True to skip checking duplicated values.
         """
         values = [values] if isinstance(values, str) else values
@@ -617,7 +639,7 @@ class Trie(StsDict):
             list_ += [x for x in values if x not in list_]
         return self
 
-    def match(self, parts, pos):
+    def match(self, parts, pos, maxlen=math.inf):
         """Match a unicode composite at pos.
 
         Args:
@@ -631,7 +653,7 @@ class Trie(StsDict):
                 list(parts))
         trie = self
         i = pos
-        total = len(parts)
+        total = min(len(parts), maxlen)
         match = None
         match_end = None
         while i < total:
@@ -651,7 +673,7 @@ class Trie(StsDict):
 class StsListMaker():
     """A class for compling a dictionary.
     """
-    def make(self, config, output_dir=None, quiet=False):
+    def make(self, config_name, base_dir=None, output_dir=None, skip_requires=False, quiet=False):
         """Compile a dictionary according to config.
 
         Load dictionaries specified in config and generate a new dictionary.
@@ -659,82 +681,68 @@ class StsListMaker():
         scheme of config (.json):
         {
             "name": "...",
+            "requires": [
+                "..."  // a required config (relative to this config file)
+            ],
             "dicts": [{
-                "file": "...",  // relative to config directory
-                "type": "...", // list, jlist, tlist
-                "mode": "...", // load, swap, join
-                "src": ["...", ...] // list of .txt or .list files
+                "file": "...",  // relative to default config directory
+                "type": "...",  // list, jlist, tlist
+                "mode": "...",  // load, swap, join
+                "src": ["...", ...]  // list of .txt or .list files
             }, ...]
         }
         """
-        def get_config_file(config):
+        def get_config_file(config, base_dir=None):
             """Calculate the path of a config file.
 
             1. Use it if it's an absolute path.
-            2. Assume relative to CWD. (.json omissible)
-            3. Assume relative to config file. (.json omissible)
+            2. Assume relative to base_dir or CWD.
+            3. Assume relative to default config directory. (.json omissible)
             4. If not found, assume relative to CWD.
             """
             if os.path.isabs(config):
                 return config
 
-            if os.path.isfile(config):
-                return config
-            for file in os.listdir():
-                filebase, fileext = os.path.splitext(file)
-                if filebase == config and fileext.lower() == '.json':
-                    return file
+            relative_config = os.path.join(base_dir, config) if base_dir is not None else config
+            if os.path.isfile(relative_config):
+                return relative_config
 
-            search_dir = os.path.normpath(os.path.join(__file__, '..', 'data', 'config'))
-            search_file = os.path.join(search_dir, config)
+            search_file = os.path.join(self.DEFAULT_CONFIG_DIR, config)
             if os.path.isfile(search_file):
                 return search_file
-            for file in os.listdir(search_dir):
-                filebase, fileext = os.path.splitext(file)
-                if filebase == config and fileext.lower() == '.json':
-                    return os.path.join(search_dir, file)
+            if not config.lower().endswith('.json'):
+                for file in glob.iglob(glob.escape(search_file) + '.[jJ][sS][oO][nN]'):
+                    return file
 
-            return config
+            return relative_config
 
-        def get_stsdict_path(stsdict):
+        def get_stsdict_file(stsdict):
             """Calculate the path of a dictionary file.
 
             1. Use it if it's an absolute path.
-            2. Assume relative to config file directory. (.json omissible)
-            3. Assume relative to dictionary directory. (.json omissible)
-            4. If not found, assume relative to config file directory.
+            2. Assume relative to the config file.
+            3. Assume relative to default dictionary directory.
+            4. If not found, assume relative to the config file.
             """
             if os.path.isabs(stsdict):
                 return stsdict
 
-            search_path = os.path.abspath(os.path.join(config_dir, stsdict))
-            if os.path.exists(search_path):
-                return search_path
+            search_file = os.path.join(config_dir, stsdict)
+            if os.path.isfile(search_file):
+                return search_file
 
-            search_path2 = os.path.abspath(os.path.join(dictionary_dir, stsdict))
-            if os.path.exists(search_path2):
-                return search_path2
+            search_file2 = os.path.join(self.DEFAULT_DICTIONARY_DIR, stsdict)
+            if os.path.isfile(search_file2):
+                return search_file2
 
-            return search_path
-
-        # Load config
-        config_file = get_config_file(config)
-        if not os.path.isfile(config_file):
-            sys.exit(f'[error] specified config "{config}" is not a file')
-
-        config_dir = os.path.abspath(os.path.dirname(config_file))
-        dictionary_dir = os.path.abspath(os.path.join(__file__, '..', 'data', 'dictionary'))
-
-        with open(config_file, "r", encoding="UTF-8") as f:
-            config = json.load(f)
-            f.close()
+            return search_file
 
         def check_update(output, filegroups):
             if not os.path.isfile(output):
                 return True
 
             for files in filegroups:
-                if type(files) is str:
+                if isinstance(files, str):
                     files = [files]
                 for file in files:
                     if os.path.getmtime(file) > os.path.getmtime(output):
@@ -742,12 +750,26 @@ class StsListMaker():
 
             return False
 
+        # locate and load the config file
+        config_file = get_config_file(config_name, base_dir)
+        config_dir = os.path.abspath(os.path.dirname(config_file))
+
+        with open(config_file, "r", encoding="UTF-8") as f:
+            config = json.load(f)
+            f.close()
+
+        # handle required configs
+        if not skip_requires:
+            for cf in config.get('requires', []):
+                self.make(cf, base_dir=config_dir, output_dir=output_dir, skip_requires=skip_requires, quiet=quiet)
+
+        # make the requested dicts
         for dict_ in config['dicts']:
             dest = os.path.join(output_dir or config_dir, dict_['file'])
             format = dict_['format']
             mode = dict_['mode']
-            files = [get_stsdict_path(f) if type(f) is str
-                else [get_stsdict_path(i) for i in f]
+            files = [get_stsdict_file(f) if isinstance(f, str)
+                else [get_stsdict_file(i) for i in f]
                 for f in dict_['src']]
 
             if not check_update(dest, files):
@@ -761,7 +783,7 @@ class StsListMaker():
             elif mode == 'swap':
                 table = Table().load(*files).swap()
             elif mode == 'join':
-                table = Table().load_fileslist(files)
+                table = Table().load_filegroups(files)
             else:
                 raise ValueError(f'Specified mode "{mode}" is not supported.')
 
@@ -778,6 +800,9 @@ class StsListMaker():
 
         return dest
 
+    DEFAULT_CONFIG_DIR = os.path.join(os.path.dirname(__file__), 'data', 'config')
+    DEFAULT_DICTIONARY_DIR = os.path.join(os.path.dirname(__file__), 'data', 'dictionary')
+
 class StsConverter():
     """Convert a text using a listfile.
     """
@@ -844,17 +869,10 @@ class StsConverter():
                     old, news = part
 
                     if self.options['mark']:
-                        content = old + "->" + "|".join(news)
-
-                        # plural > exact > single
-                        if len(news) > 1:
-                            part = "{{" + content + "}}"
-                        elif old == news[0]:
-                            part = "{{" + content + "}}"
-                        elif len(news) == 1:
-                            part = "{{" + content + "}}"
+                        if len(news) == 1 and old == news[0]:
+                            part = "{{" + old + "}}"
                         else:
-                            part = news[0]
+                            part = "{{" + old + "->" + "|".join(news) + "}}"
                     else:
                         part = news[0]
 
@@ -870,11 +888,11 @@ class StsConverter():
                     for i, v in enumerate(news):
                         content += f'<ins{" hidden" if i else ""}>{html.escape(v)}</ins>'
 
-                    # plural > exact > single
-                    classes = []
+                    # classes
+                    classes = ['sts-conv']
                     if len(news) == 1:
                         classes.append('single')
-                    if len(news) > 1:
+                    else:
                         classes.append('plural')
                     if old == news[0]:
                         classes.append('exact')
@@ -969,7 +987,7 @@ def main():
     def make(args):
         """Compile conversion dictionary(s).
         """
-        configs = ['_default'] + args['config']
+        configs = args['config']
         dir = args['dir']
         quiet = args['quiet']
 
@@ -988,7 +1006,6 @@ def main():
             'exclude': args['exclude'],
             }
 
-        StsListMaker().make('_default', quiet=True)
         stsdict = StsListMaker().make(config, quiet=True)
         converter = StsConverter(stsdict, options)
         converter.convert_file(input, output)
